module(..., package.seeall)

--------------------------------------------------------------------------------------
--																					--
-- Description: This module provides all needed function for a basic pathfinding	--
-- 				algo, more precisly the A* search algorithm is implemented			--
--																					--
-- Author: drs																		--
--																					--
-- Sources: http://en.wikipedia.org/wiki/A*_search_algorithm						--
--			https://github.com/lattejed/a-star-lua/blob/master/a-star.lua			--
--			http://www.ohloh.net/p/A-Star-Pathfinding-For-Lua						--
--																					--
-- Version: 0.1.0																	--
--------------------------------------------------------------------------------------

----------------------------------------------------------------------------------------------------------
-- function: 		calcPath(startPos, endPos, mode)
-- description:		creates a track from start to end using a* algo
-- parameter:		startPos; endPos
--					mode: 	"avoidAll" 		-> recognize and take all entities and water tiles into calculation
--							"avoidWater"	-> just avoid watertiles
--							nil				-> ignores all entities/ tiles; using a complete different algo instead of this option might be the better choice, in terms of performance
-- return:			closedList if finished;
--					currentNode if not finished yet
--					nil if target postion cannot be reached
----------------------------------------------------------------------------------------------------------
function calcPath(startPos, endPos, mode)
	
	local tempG, tempH = 0, 0

	if openlist == nil then
		openlist = {}
		closedlist = {}
		tempG = 0
		tempH = calcH(startPos.x, startPos.y, endPos)
		-- place startNode into openlist
		openlist[1] = {x = startPos.x, y = startPos.y, G = tempG, H = tempH, F = tempG + tempH , parent = 1}
	end

	-- run the if-then each function-call until openlist is empty
	if #openlist > 0 then
		currentNode = calcLowestF(openlist)
		
		-- target reached?
		if (currentNode.x == endPos.x) and (currentNode.y == endPos.y) then
			return retPath(closedlist)
		end

		-- remove current node from openlist and put it into closedlist
		removeNode (openlist, currentNode)                
		table.insert (closedlist, currentNode)

		-- if target not reached => get neighbours of currentNode
		local neighbours = findNeighbours(currentNode, closedlist, mode)
		-- calculate G-value of neighbours
		for _, nodes in ipairs(neighbours) do
			tempG = currentNode.G + 1
			if not inList(nodes.x, nodes.y, openlist) then
				-- node is not present in openlist
				tempH = calcH(nodes.x, nodes.y, endPos)
				table.insert(openlist,{x = nodes.x, y = nodes.y, G = tempG, H = tempH, F = tempG + tempH, parent = #closedlist})
			else
				-- node is present in openlist
				for _, openlistNodes in ipairs(openlist) do
					if (nodes.x == openlistNodes.x) and (nodes.y == openlistNodes.y) then
						if tempG < openlistNodes.G then
							openlistNodes.G = tempG
							openlistNodes.H = calcH(nodes.x, nodes.y, endPos)
							openlistNodes.F = tempG + openlistNodes.H
							openlistNodes.parent = #closedlist
						end
						break
					end
				end
			end		
		end
	end
	-- openlist is empty =>no path was found
	if (#openlist == 0) then
		-- reset all tables and return nil on error
		openlist = nil
		closedlist = nil
		currentNode = nil
		return nil
	end
	-- return currentNode if not done yet
	return currentNode
end

----------------------------------------------------------------------------------------------------------
-- function: 		calcH(Nx, Ny, endPos)
-- description:		calculate the H-costs of the current node (manhatten-distance)
-- parameter:		currentNodePos; endPos
-- return:			H-cost of current node
----------------------------------------------------------------------------------------------------------
function calcH(Nx, Ny, endPos)
	return (math.abs(Nx - endPos.x) + math.abs(Ny - endPos.y)) * 10
end

----------------------------------------------------------------------------------------------------------
-- function: 		calcLowestF(openlist)
-- description:		look trough openlist and returns the node with the lowest F-score
-- parameter:		openlist
-- return:			node with lowest F-score
----------------------------------------------------------------------------------------------------------
function calcLowestF(openlist)

	local lowestF = 1/0 -- initialize the f-score with infinity
	local bestNode = nil

	for _, node in ipairs (openlist) do
		if node.F < lowestF then
			lowestF = node.F
			bestNode = node
		end
	end

	return bestNode 
end

----------------------------------------------------------------------------------------------------------
-- function: 		removeNode(openlist, currentNode)
-- description:		removes the passed node from the openlist
-- parameter:		openlist, currentNode
-- return:			none
----------------------------------------------------------------------------------------------------------
function removeNode(openlist, currentNode)
	for i, node in ipairs (openlist) do
		if node == currentNode then
			table.remove(openlist, i)
			break
		end
	end        
end

----------------------------------------------------------------------------------------------------------
-- function: 		findNeighbours(thisNode, closedlist, mode)
-- description:		finds and returns the neighbours of the passed node
--					(that are not already in the closedlist)
-- parameter:		thisNode, closedlist, mode
-- return:			neighbours
----------------------------------------------------------------------------------------------------------
function findNeighbours(thisNode, closedlist, mode)
	local neighbours = {}
	-- neighbour @ top
	if not inList(thisNode.x, thisNode.y - 1, closedlist) then
		if mode == "avoidAll" then
			if (game.canplaceentity{name="wooden-chest", position = {x = thisNode.x, y = thisNode.y - 1}}) then
				table.insert(neighbours,{x = thisNode.x, y = thisNode.y - 1})
			end
		elseif mode == "avoidWater" then
			local tileName = game.gettile(thisNode.x,thisNode.y - 1).name
			if tileName ~= "out-of-map" and tileName ~= "deepwater" and tileName ~= "deepwater-green" and tileName ~= "water" and tileName ~= "water-green" then 
				table.insert(neighbours,{x = thisNode.x, y = thisNode.y - 1})
			end
		else
			table.insert(neighbours,{x = thisNode.x, y = thisNode.y - 1})
		end
	end	
	-- neighbour @ right
	if not inList(thisNode.x + 1, thisNode.y, closedlist) then
		if mode == "avoidAll" then
			if (game.canplaceentity{name="wooden-chest", position = {x = thisNode.x + 1, y = thisNode.y}}) then
				table.insert(neighbours,{x = thisNode.x + 1, y = thisNode.y})
			end
		elseif mode == "avoidWater" then
			local tileName = game.gettile(thisNode.x + 1,thisNode.y).name
			if tileName ~= "out-of-map" and tileName ~= "deepwater" and tileName ~= "deepwater-green" and tileName ~= "water" and tileName ~= "water-green" then 
				table.insert(neighbours,{x = thisNode.x + 1, y = thisNode.y})
			end
		else
			table.insert(neighbours,{x = thisNode.x + 1, y = thisNode.y})
		end
	end
	-- neighbour @ bottom
	if not inList(thisNode.x, thisNode.y + 1, closedlist) then
		if mode == "avoidAll" then
			if (game.canplaceentity{name="wooden-chest", position = {x = thisNode.x, y = thisNode.y + 1}}) then
				table.insert(neighbours,{x = thisNode.x, y = thisNode.y + 1})
			end
		elseif mode == "avoidWater" then
			local tileName = game.gettile(thisNode.x,thisNode.y + 1).name
			if tileName ~= "out-of-map" and tileName ~= "deepwater" and tileName ~= "deepwater-green" and tileName ~= "water" and tileName ~= "water-green" then
				table.insert(neighbours,{x = thisNode.x, y = thisNode.y + 1})
			end
		else
			table.insert(neighbours,{x = thisNode.x, y = thisNode.y + 1})
		end
	end
	-- neighbour @ left
	if not inList(thisNode.x - 1, thisNode.y, closedlist) then
		if mode == "avoidAll" then
			if (game.canplaceentity{name="wooden-chest", position = {x = thisNode.x - 1, y = thisNode.y}}) then
				table.insert(neighbours,{x = thisNode.x - 1, y = thisNode.y})
			end
		elseif mode == "avoidWater" then
			local tileName = game.gettile(thisNode.x - 1,thisNode.y).name
			if tileName ~= "out-of-map" and tileName ~= "deepwater" and tileName ~= "deepwater-green" and tileName ~= "water" and tileName ~= "water-green" then
				table.insert(neighbours,{x = thisNode.x - 1, y = thisNode.y})
			end
		else
			table.insert(neighbours,{x = thisNode.x - 1, y = thisNode.y})
		end
	end

	return neighbours 
end

----------------------------------------------------------------------------------------------------------
-- function: 		inList(thisNode, list)
-- description:		checks if the passed node is in the passed list
-- parameter:		thisNode, list
-- return:			true if in list; false if not in list
----------------------------------------------------------------------------------------------------------
function inList(Nx, Ny, list)
	local thisNode = {x = Nx, y = Ny}
	for _, listNodes in ipairs(list) do
		if (listNodes.x == thisNode.x) and (listNodes.y == thisNode.y) then
			return true
		end
	end
	return false
end
----------------------------------------------------------------------------------------------------------
-- function: 		retPath(closedlist)
-- description:		find and extracts the path from the passed closedlist
-- parameter:		closedlist
-- return:			path as a list with positions
----------------------------------------------------------------------------------------------------------
function retPath(closedlist)

	if closedlist == nil then
		return nil
	end

	local path = {}
	local pathIndex = {}
	local last = #closedlist
	table.insert(pathIndex, last)

	local i = 1
	while pathIndex[i] > 1 do
		i = i + 1
		table.insert(pathIndex, i, closedlist[pathIndex[i - 1]].parent)
	end

	for n = #pathIndex, 1, -1 do
		table.insert(path, {x = closedlist[pathIndex[n]].x, y = closedlist[pathIndex[n]].y})
	end

	closedlist=nil
	return path
end

----------------------------------------------------------------------------------------------------------
-- function: 		resetA_StarTables()
-- description:		used to reset all related tables after a path was found
--					on pathfind error this is done automously, so there is no need to call this function then
-- parameter:		none
-- return:			none
----------------------------------------------------------------------------------------------------------
function resetA_StarTables()
	openlist = nil
	closedlist = nil
	currentNode = nil
end